##
#
# File:    testChemCompModelBuild.py
# Author:  J. Westbrook
# Date:    19-Jan-2021
# Version: 0.001
#
# Updated:
#
##
"""
Test cases for to build models from CCDC search results in the chemical component model workflow.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import unittest
import time
import os
import os.path
import platform
import resource

from rcsb.ccmodels.search.ChemCompModelBuild import ChemCompModelBuild

from rcsb.ccmodels.search import __version__


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ChemCompModelBuildTests(unittest.TestCase):
    def setUp(self):
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__prefix = "abbrev"
        self.__numProc = 2
        self.__chunkSize = 5
        #
        self.__startTime = time.time()
        logger.info("Starting %s (%s) at %s", self.id(), __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testBuildWorkflow(self):
        """Test case: model build workflow step"""
        try:
            ccmb = ChemCompModelBuild(cachePath=self.__cachePath, prefix=self.__prefix)
            rD = ccmb.build(alignType="graph-relaxed-stereo-sdeq", numProc=2, chunkSize=2)
            logger.info("Matched search ids %r", list(rD.keys()))
            self.assertGreaterEqual(len(rD), 9)
            qD = ccmb.fetchModelIndex()
            self.assertEqual(len(rD), len(qD))
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def suiteBuildTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ChemCompModelBuildTests("testBuildWorkflow"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite = suiteBuildTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    ChemCompModelAssemble.py
# Author:  J. Westbrook
# Date:     4-Feb-2021
# Version: 0.001
#
# Updated:
##
"""
Assemble model files and adjust audit records for the chemical component model workflow.
Models identifiers and audit creation dates are reconciled with the audit details of the
prior public model collection.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import datetime
import logging
import os
import time
from collections import defaultdict
from operator import itemgetter

from rcsb.ccmodels.search.ChemCompModelBuild import ChemCompModelBuild
from rcsb.ccmodels.search import __version__
from rcsb.utils.chemref.ChemCompModelProvider import ChemCompModelProvider
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompModelAssemble(object):
    def __init__(self, cachePath, prefix=None, urlTarget=None):
        self.__cachePath = cachePath
        self.__prefix = prefix
        self.__urlTarget = urlTarget
        #
        self.__ccmb = ChemCompModelBuild(cachePath=self.__cachePath, prefix=self.__prefix)
        # self.__startTime = time.time()
        logger.info("Starting assemble (%s) at %s", __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def assemble(self, maxRFactor=10.0):
        """Concatenate models into the input file path subject to the R value constraint.
        Relabel the models sequentially for each parent chemical component.

        Args:
            assembleModelPath (str): path for concatenated model file
            maxRFactor (float, optional): limiting R-value. Defaults to 10.0.

        Returns:
            (bool): True for success or False otherwise

        """
        dataContainerL = []
        mU = MarshalUtil(workPath=self.__cachePath)
        modelIndexD = self.__ccmb.fetchModelIndex()
        modelIndexD = self.__addPriorMatchDetails(modelIndexD)
        modelIndexD = self.__updateVariantDetails(modelIndexD)
        priorMapD = {}
        for _, mDL in modelIndexD.items():
            mDLS = sorted(mDL, key=itemgetter("priorModelId", "variantType", "rFactor"), reverse=False)

            numStd = 0
            matchIdD = {}
            for mD in mDLS:
                isStd = False
                if mD["variantType"].startswith("A"):
                    numStd += 1
                    isStd = True
                #
                if mD["rFactor"] > maxRFactor:
                    logger.info("Skipping model %s isStd (%r) rValue (%r)", mD["modelId"], isStd, mD["rFactor"])
                    continue
                if numStd and not isStd:
                    logger.info("Skipping model %s isStd (%r) numStd (%d)", mD["modelId"], isStd, numStd)
                    continue
                #
                # Exclude duplicate matches in priority order ...
                if mD["matchId"] in matchIdD:
                    logger.info("Skipping duplicate matchId %r in %r", mD["matchId"], mD["modelId"])
                    continue
                #
                matchIdD[mD["matchId"]] = True

                cL = mU.doImport(mD["modelPath"], fmt="mmcif")
                logger.debug("Read %d from %s", len(cL), mD["modelPath"])
                dataContainerL.extend(cL)
                if not mD["priorModelId"].startswith("Z"):
                    priorMapD[mD["modelId"]] = (mD["priorModelId"], mD["priorMatchDate"])
        #
        logger.debug("priorMapD %r", priorMapD)
        fn = "chem_comp_models-%s.cif" % self.__getToday()
        assembleModelPath = os.path.join(self.__ccmb.getModelDirFilePath(), fn)
        # -- relabel
        parentModelCountD = defaultdict(int)
        priorIdLD = {}
        for dataContainer in dataContainerL:
            tModelId = dataContainer.getName()
            tId = tModelId.split("_")[1]
            pId = tId.split("|")[0]
            # tCount = tModelId.split("_")[2]
            if tModelId in priorMapD:
                pCount = priorMapD[tModelId][0].split("_")[2]
                priorIdLD.setdefault(pId, []).append(int(pCount))
                self.__replaceModelId(dataContainer, tModelId, priorMapD[tModelId][0])
                self.__updateAuditDate(dataContainer, priorMapD[tModelId][1])
                parentModelCountD[pId] = sorted(priorIdLD[pId])[-1]
                logger.debug("%s current model %r prior model %r count %d", pId, tModelId, priorMapD[tModelId][0], parentModelCountD[pId])
            else:
                parentModelCountD[pId] += 1
                pModelId = self.__makePublicModelId(pId, parentModelCountD[pId])
                self.__replaceModelId(dataContainer, tModelId, pModelId)

        ok = mU.doExport(assembleModelPath, dataContainerL, fmt="mmcif")
        logger.info("Assembled %d models status %r", len(dataContainerL), ok)
        self.__checkAssembledModels(assembleModelPath)
        return len(dataContainerL)

    def __getAuditDetails(self):
        """[summary]

        Returns:
            [type]: [description]

             aL.append({"audit_date": auditDate, "action_type": auditAction})
             rD.setdefault(ccId, []).append({"model_id": modelId, "db_name": dbName, "db_code": dbCode, "audit_list": aL})
        """
        if self.__urlTarget:
            ccm = ChemCompModelProvider(cachePath=self.__cachePath, useCache=False, urlTarget=self.__urlTarget)
        else:
            ccm = ChemCompModelProvider(cachePath=self.__cachePath, useCache=False)
        rD = ccm.getAuditDetails()
        return rD

    def __addPriorMatchDetails(self, modelIndexD):
        """"""
        priorMatchLD = self.__getAuditDetails()
        for pId, mDL in modelIndexD.items():
            for mD in mDL:
                mD["priorModelId"] = "Znone"
            if pId in priorMatchLD:
                logger.debug("priorMatchLD %r", priorMatchLD[pId])
                numMatch = 0
                for priorMatch in priorMatchLD[pId]:
                    priorModelId = priorMatch["model_id"]
                    priorDbName = priorMatch["db_name"]
                    priorDbCode = priorMatch["db_code"]
                    priorDate = priorMatch["audit_list"][-1]["audit_date"]
                    # compare with current matches
                    for mD in mDL:
                        curDbName = "CSD"
                        # curDbName = mD["matchDb"]
                        curDbCode = mD["matchId"]
                        if priorDbName == curDbName and curDbCode == priorDbCode:
                            numMatch += 1
                            mD["priorModelId"] = priorModelId
                            mD["priorMatchDate"] = priorDate
                if numMatch:
                    logger.info("%s has prior matches (%d)", pId, numMatch)
        return modelIndexD

    def __updateVariantDetails(self, modelIndexD):
        """"""
        for _, mDL in modelIndexD.items():
            for mD in mDL:
                if not mD["variantType"]:
                    mD["variantType"] = "Anone"
        return modelIndexD

    def __makePublicModelId(self, parentId, modelNum=1):
        modelId = "M_" + parentId + "_%05d" % modelNum
        return modelId

    def __checkAssembledModels(self, assembleModelPath):
        catNameL = [
            "pdbx_chem_comp_model",
            "pdbx_chem_comp_model_atom",
            "pdbx_chem_comp_model_bond",
            "pdbx_chem_comp_model_descriptor",
            "pdbx_chem_comp_model_reference",
            "pdbx_chem_comp_model_feature",
            "pdbx_chem_comp_model_audit",
        ]
        mU = MarshalUtil(workPath=self.__cachePath)
        dataContainerL = mU.doImport(assembleModelPath, fmt="mmcif")
        logger.info("Read %d data containers", len(dataContainerL))
        rD = {}
        cnD = {}
        for dataContainer in dataContainerL:
            nm = dataContainer.getName()
            logger.debug("datacontainer %r", nm)
            if nm in cnD:
                logger.info("Duplicate container id %r", nm)
                cnD[nm] = True
            #
            pId = nm.split("_")[1]
            cObj = dataContainer.getObj("pdbx_chem_comp_model")
            modelId = cObj.getValue("id", 0)
            if modelId != nm:
                logger.error("modelId %r datablock %r", modelId, nm)
            #
            tD = {}
            for catName in catNameL:
                cObj = dataContainer.getObj(catName)
                nRows = cObj.getRowCount()
                tD[catName] = nRows
            cObj = dataContainer.getObj("pdbx_chem_comp_model_feature")
            skip = False
            for ii in range(cObj.getRowCount()):
                fN = cObj.getValue("feature_name", ii)
                fV = cObj.getValue("feature_value", ii)
                if fN == "heavy_atoms_only" and fV == "Y":
                    skip = True
                    break
            if not skip:
                rD.setdefault(pId, []).append(tD)
        #
        for pId, tDL in rD.items():
            for catName in catNameL:
                minV = 100000
                maxV = -1
                for tD in tDL:
                    minV = min(minV, tD[catName])
                    maxV = max(maxV, tD[catName])
                if maxV - minV > 1 and catName not in ["pdbx_chem_comp_model_feature"]:
                    logger.error("%s %s row count inconsistency %d %d", pId, catName, minV, maxV)

    def __replaceModelId(self, dataContainer, oldModelId, newModelId):
        """Update all instances of the model in the input container with the input
        replacement value.

        Args:
            dataContainerList (obj): input container list
            newModelId (str):  replacement modelId value

        """
        updateL = [
            ("pdbx_chem_comp_model", "id"),
            ("pdbx_chem_comp_model_atom", "model_id"),
            ("pdbx_chem_comp_model_bond", "model_id"),
            ("pdbx_chem_comp_model_descriptor", "model_id"),
            ("pdbx_chem_comp_model_reference", "model_id"),
            ("pdbx_chem_comp_model_feature", "model_id"),
            ("pdbx_chem_comp_model_audit", "model_id"),
        ]
        tup = ()
        try:
            dataContainer.setName(newModelId)
            for tup in updateL:
                cObj = dataContainer.getObj(tup[0])
                cObj.replaceValue(oldModelId, newModelId, tup[1])
        except Exception as e:
            logger.exception("Failing for cat %r %r and %r with %s", tup, oldModelId, newModelId, str(e))
        return dataContainer

    def __updateAuditDate(self, dataContainer, auditDate):
        """Update the audit date for input container

        Args:
            dataContainerList (obj): input container list
            auditDate (str):  original audit date

        """
        try:
            cObj = dataContainer.getObj("pdbx_chem_comp_model_audit")
            ii = cObj.getRowCount()
            cObj.setValue(auditDate, "date", ii - 1)
        except Exception as e:
            logger.exception("Failing for %r with %s", auditDate, str(e))
        return dataContainer

    def __getToday(self):
        """Return a CIF style date-timestamp value for current local time -"""
        today = datetime.datetime.today()
        # format ="%Y-%m-%d:%H:%M"
        fmt = "%Y-%m-%d"
        return str(today.strftime(fmt))

    def __makePublicModelId(self, parentId, modelNum=1):
        modelId = "M_" + parentId + "_%05d" % modelNum
        return modelId



##
# File:    ChemCompModelBuild.py
# Author:  J. Westbrook
# Date:    19-Jan-2021
# Version: 0.001
#
# Updated:
##
"""
Select and build model files for chemical component model workflow.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import copy
import datetime
import logging
import os
import time
from collections import namedtuple, defaultdict

from mmcif.api.DataCategory import DataCategory
from mmcif.api.PdbxContainers import DataContainer
from rcsb.ccmodels.search import __version__
from rcsb.ccmodels.search.ChemCompModelSearch import ChemCompModelSearch
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.OeAlignUtils import OeAlignUtils
from rcsb.utils.chem.OeDepictAlign import OeDepictMCSAlignMultiPage
from rcsb.utils.chem.OeDepictAlign import OeDepictMCSAlignPage
from rcsb.utils.io.IndexUtils import CcdcMatchIndex
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)

ComponentAtomDetails = namedtuple("ComponentAtomDetails", "atIdx atNo atName atType x y z atFormalCharge")
AlignAtomMap = namedtuple("AlignAtomMap", "refId refAtIdx refAtNo refAtName fitId fitAtIdx fitAtNo fitAtName")
AlignAtomUnMapped = namedtuple("AlignAtomUnMapped", "fitId fitAtIdx fitAtNo fitAtType fitAtName fitAtFormalCharge x y z fitNeighbors")


class ChemCompModelBuildWorker(object):
    def __init__(self, cachePath, verbose=True):
        self.__cachePath = cachePath
        self.__verbose = verbose
        self.__ccSIdxP = None

    def build(self, dataList, procName, optionsD, workingDir):
        """Worker method to build chemical component model for the input search result path list.

        Args:
            dataList (list): list of search result paths
            procName (str): processName
            optionsD (dict): dictionary of options
            workingDir (str): path to working directory (not used)

        Returns:
            (successList, resultList, []): success and result lists
        """
        _ = workingDir
        resultList = []
        successList = []
        #
        try:
            alignType = optionsD["alignType"]
            modelDirPath = optionsD["modelDirPath"]
            imageDirPath = optionsD["imageDirPath"]
            self.__ccSIdxP = optionsD["ccSIdxP"]
            #
            startTime = time.time()
            logger.info("%s ======== ============ ============ ", procName)
            logger.info("%s starting with length %d at %s", procName, len(dataList), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
            #
            parentD = {}
            parentModelCountD = defaultdict(int)
            for idxPath in dataList:
                fn = os.path.basename(idxPath)
                sId = fn.replace("-index.json", "")
                parentId = sId.split("|")[0]
                logger.info("%s start model build for search Id (%s) %s", procName, parentId, sId)
                #
                matchObjIt = CcdcMatchIndex(indexFilePath=idxPath)
                matchObjIt.sort()
                pairList = []

                for matchObj in matchObjIt:
                    targetCifPath = matchObj.getTargetCcPath()
                    fitMolFilePath = matchObj.getMol2Path()
                    matchId = matchObj.getIdentifier()
                    targetId = matchObj.getTargetId()
                    matchTitle = "CCDC Code  " + matchId
                    ccTitle = "Chemical Component " + targetId
                    #
                    nAtomsRef, refFD, nAtomsFit, fitFD, fitXyzMapD, fitAtomUnMappedL, isSkipped = self.__alignModelSubStruct(
                        targetCifPath,
                        fitMolFilePath,
                        alignType=alignType,
                        fitTitle=matchId,
                        refTitle=targetId,
                        onlyCloseMatches=True,
                        verbose=self.__verbose,
                        procName=procName,
                    )
                    logger.debug(
                        ">>> %s - %s nAtomsRef %d nAtomsFit %d atommapL (%d) fitAtomUnMappedL (%d)", targetId, matchId, nAtomsRef, nAtomsFit, len(fitXyzMapD), len(fitAtomUnMappedL)
                    )
                    if nAtomsRef == 0 and nAtomsFit == 0:
                        logger.info("%s alignment fails for %s with %s", procName, targetId, matchId)
                        continue
                    if isSkipped:
                        logger.info("%s alignment skipped for %s (%d) with %s (%d)", procName, targetId, nAtomsRef, matchId, nAtomsFit)
                        continue

                    # -----
                    smilesMatch = refFD["SMILES_STEREO"] == fitFD["SMILES_STEREO"]
                    hasUnMapped = len(fitAtomUnMappedL) > 0
                    unMappedOk = self.__testUnMappedProtonation(fitAtomUnMappedL)
                    #
                    #   --- acceptance tests ---
                    acceptOk = False
                    if nAtomsRef == len(fitXyzMapD) and smilesMatch:
                        # Take all cases with matching SMILES and all atoms aligned
                        acceptOk = True
                    elif nAtomsRef > len(fitXyzMapD) and smilesMatch:
                        # Heavy atom only match
                        acceptOk = True
                    elif nAtomsRef == len(fitXyzMapD) and hasUnMapped and unMappedOk:
                        # All atoms aligned and fit includes a protonation ...
                        acceptOk = True
                    else:
                        pass

                    if len(fitXyzMapD) < 2:
                        acceptOk = False
                    #
                    if not acceptOk:
                        logger.info(
                            "%s rejecting (smilesMatch %r hasUnmapped %r (%d) unMappedOk %r nAtomsRef %d nAtomsFit %d mapped fit %d) for %s with %s",
                            procName,
                            smilesMatch,
                            hasUnMapped,
                            len(fitAtomUnMappedL),
                            unMappedOk,
                            nAtomsRef,
                            nAtomsFit,
                            len(fitXyzMapD),
                            targetId,
                            matchId,
                        )
                        continue
                    #
                    if hasUnMapped and not smilesMatch:
                        logger.info("%s SMILES for %s and %s differ with unmapped protons", procName, targetId, matchId)
                        logger.debug("Ref %-8s SMILES: %s", targetId, refFD["SMILES_STEREO"])
                        logger.debug("Fit %-8s SMILES: %s", matchId, fitFD["SMILES_STEREO"])

                    # --------- ----------------
                    #  Accept the match
                    # --------- ----------------
                    matchId = matchObj.getIdentifier()
                    targetId = matchObj.getTargetId()
                    parentD.setdefault(parentId, []).append(matchId)
                    #
                    refImageFileName = "ref_" + targetId + "_" + matchId + ".svg"
                    refImagePath = os.path.join(imageDirPath, sId, refImageFileName)
                    self.__pairDepictPage(refImagePath, sId, ccTitle, refFD["OEMOL"], matchId, matchTitle, fitFD["OEMOL"], alignType=alignType)
                    # --------- ------------------
                    pairList.append((sId, refFD["OEMOL"], matchId, fitFD["OEMOL"]))
                    modelId, modelPath = self.__makeModelPath(modelDirPath, parentId, targetId, startingModelNum=parentModelCountD[parentId] + 1, maxModels=300, scanExisting=False)
                    logger.debug("targetId %r modelId %r modelPath %r", targetId, modelId, modelPath)
                    #
                    logger.info(
                        "%s accepted for %s with %s (smilesMatch %r hasUnmapped %r (%d) unMappedOk %r nAtomsRef %d nAtomsFit %d mapped fit %d) ",
                        procName,
                        targetId,
                        matchId,
                        smilesMatch,
                        hasUnMapped,
                        len(fitAtomUnMappedL),
                        unMappedOk,
                        nAtomsRef,
                        nAtomsFit,
                        len(fitXyzMapD),
                    )
                    ok, variantType = self.__writeModel(targetId, targetCifPath, fitFD, fitXyzMapD, fitAtomUnMappedL, matchObj, modelId, modelPath)
                    if ok:
                        parentModelCountD[parentId] += 1
                        hd = matchObj.getHasDisorder()
                        resultList.append(
                            {
                                "modelId": modelId,
                                "searchId": targetId,
                                "parentId": parentId,
                                "matchId": matchId,
                                "matchDB": "CSD",
                                "modelPath": modelPath,
                                "rFactor": matchObj.getRFactor(),
                                "hasDisorder": hd if hd else "N",
                                "variantType": variantType,
                            }
                        )
                #
                if pairList:
                    pdfImagePath = os.path.join(imageDirPath, sId, sId + "-all-pairs.pdf")
                    self.__depictFitList(sId, pdfImagePath, pairList, alignType=alignType)
                if resultList:
                    logger.info("%s built %d models for %s (this iteration)", procName, parentModelCountD[parentId], parentId)
                    successList.append(idxPath)
                else:
                    logger.info("%s no models built for %s", procName, sId)

            endTime = time.time()
            logger.info(
                "%s (match successes %d total models this iterations %d) completed at %s (%.2f seconds)",
                procName,
                len(successList),
                len(resultList),
                time.strftime("%Y %m %d %H:%M:%S", time.localtime()),
                endTime - startTime,
            )
            return successList, resultList, []
        except Exception as e:
            logger.exception("%s failing with %s", procName, str(e))
        return [], [], []

    def __getBuildVariant(self, sId):
        """Lookup the build type from the input chemical component search index Id.

        Args:
            sId (str): chemical component search index Id

        Returns:
            str : "tautomer_protomer" | ""
        """
        sidxD = self.__ccSIdxP.getIndexEntry(sId)
        if "protomer" in sidxD["build-type"]:
            return "tautomer_protomer"
        if "tautomer" in sidxD["build-type"]:
            return "tautomer_protomer"
        return ""

    def __alignModelMcss(
        self, ccRefPath, molFitPath, alignType="stict", fitTitle=None, refTitle=None, unique=False, minFrac=0.9, onlyCloseMatches=False, verbose=False, procName="main"
    ):
        """Align (substructure) chemical component definition search target with the candidate matching reference molecule.

        Args:
            ccRefPath (str): path to search target chemical component definition
            molFitPath (str): path to matched molfile
            alignType (str, optional):  strict|relaxed|relaxed-stereo. Defaults to "strict"
            fitTitle (str, optional): fit molecule title. Defaults to None.
            refTitle (str, optional): reference molecule title. Defaults to None.
            unique (bool, optional): find unique matches. Defaults to False.
            minFrac (float, optional): minimum atom-level matching coverage fraction. Defaults to 0.9.
            onlyCloseMatches (bool, optional): triage alignments for only close matches. Defaults to False
            verbose (bool, optional): enable verbose output. Defaults to False.
            procName (str, optional):  process name.  Defaults to main

        Returns:
            (tuple): (  number of atoms reference molecule,
                        reference molecule feature dictionary {"Formula": , "SMILES": , "SMILES_STEREO": , "InChI": , "InChIKey":,}
                        number of atoms in fit molecule,
                        fit molecule feature dictionary  {"Formula": , "SMILES": , "SMILES_STEREO": , "InChI": , "InChIKey":, 'xyz': }
                        atomMapD {refAtName: (fit x, fit y, fit z)}
                        fitAtomUnMappedL: [(atName, atomicNumber),(), ...],
                        isSkipped
                    )

            Note: 'xyz': [ComponentAtomDetails()]
        """
        try:
            refFD = fitFD = atomMapD = {}
            fitAtomUnMappedL = []
            isSkipped = False
            logger.debug("Align target cc %s with matching model %s", ccRefPath, molFitPath)
            oesU = OeAlignUtils(workPath=self.__cachePath, verbose=verbose)
            oesU.setSearchType(sType=alignType)

            nAtomsRef = oesU.setRefPath(ccPath=ccRefPath, title=refTitle)
            nAtomsFit = oesU.setFitPath(molFitPath, title=fitTitle, suppressHydrogens=False, fType="sdf", importType="3D")
            if onlyCloseMatches and nAtomsFit > nAtomsRef and nAtomsFit - nAtomsRef > 2:
                isSkipped = True
                return nAtomsRef, refFD, nAtomsFit, fitFD, atomMapD, fitAtomUnMappedL, isSkipped
            (nAtomsRef, refFD, nAtomsFit, fitFD, atomMapL, fitAtomUnMappedL) = oesU.doAlignMcss(unique=unique, minFrac=minFrac)
            # -----
            # tD = {refAtName: fitAtIdx, ...}
            tD = {tup.refAtName: tup.fitAtIdx for tup in atomMapL}
            #
            # fitFD['xyz']: [ComponentAtomDetails()]
            # fitXyzD  = {atIdx1: ComponentAtomDetails(), atIdx2: ComponentAtomDetails(), ...}
            fitXyzD = {tup.atIdx: tup for tup in fitFD["xyz"]}
            atomMapD = {atName: fitXyzD[fitAtIdx] for atName, fitAtIdx in tD.items() if fitAtIdx in fitXyzD}
            # -----
            return nAtomsRef, refFD, nAtomsFit, fitFD, atomMapD, fitAtomUnMappedL, isSkipped
        except Exception as e:
            logger.exception("%a failing for %r and %r with %s", procName, ccRefPath, molFitPath, str(e))
        return 0, {}, 0, {}, {}, [], False

    def __alignModelSubStruct(self, ccRefPath, molFitPath, alignType="strict", fitTitle=None, refTitle=None, onlyCloseMatches=False, verbose=False, procName="main"):
        """Align (substructure) chemical component definition search target with the candidate matching reference molecule.

        Args:
            ccRefPath (str): path to search target chemical component definition
            molFitPath (str): path to matched molfile
            alignType (str, optional):  strict|relaxed|relaxed-stereo. Defaults to "strict"
            fitTitle (str, optional): fit molecule title. Defaults to None.
            refTitle (str, optional): reference molecule title. Defaults to None.
            onlyCloseMatches (bool, optional): triage alignments for only close matches. Defaults to False
            verbose (bool, optional): enable verbose output. Defaults to False.
            procName (str, optional):  process name.  Defaults to main


        Returns:
            (tuple): (  number of atoms reference molecule,
                        reference molecule feature dictionary {"Formula": , "SMILES": , "SMILES_STEREO": , "InChI": , "InChIKey":,}
                        number of atoms in fit molecule,
                        fit molecule feature dictionary  {"Formula": , "SMILES": , "SMILES_STEREO": , "InChI": , "InChIKey":, 'xyz': }
                        atomMapD {refAtName: (fit x, fit y, fit z)}
                        fitAtomUnMappedL: [(atName, atomicNumber),(), ...],
                        isSkipped
                    )

            Note: 'xyz': [ComponentAtomDetails()]
        """
        try:
            refFD = fitFD = atomMapD = {}
            fitAtomUnMappedL = []
            isSkipped = False
            #
            logger.debug("Align target cc %s with matching model %s", ccRefPath, molFitPath)
            oesU = OeAlignUtils(workPath=self.__cachePath, verbose=verbose)
            oesU.setSearchType(sType=alignType)
            nAtomsRef = oesU.setRefPath(ccPath=ccRefPath, title=refTitle)
            nAtomsFit = oesU.setFitPath(molFitPath, title=fitTitle, suppressHydrogens=False, fType="sdf", importType="3D")
            if onlyCloseMatches and nAtomsFit > nAtomsRef and nAtomsFit - nAtomsRef > 2:
                isSkipped = True
                return nAtomsRef, refFD, nAtomsFit, fitFD, atomMapD, fitAtomUnMappedL, isSkipped
            (nAtomsRef, refFD, nAtomsFit, fitFD, atomMapL, fitAtomUnMappedL) = oesU.doAlignSs(unique=False)
            # -----
            # tD = {refAtName: fitAtIdx, ...}
            tD = {tup.refAtName: tup.fitAtIdx for tup in atomMapL}
            #
            # fitFD['xyz']: [ComponentAtomDetails()]
            # fitXyzD  = {atIdx1: ComponentAtomDetails(), atIdx2: ComponentAtomDetails(), ...}
            fitXyzD = {tup.atIdx: tup for tup in fitFD["xyz"]}
            atomMapD = {refAtName: fitXyzD[fitAtIdx] for refAtName, fitAtIdx in tD.items() if fitAtIdx in fitXyzD}
            # -----
            return nAtomsRef, refFD, nAtomsFit, fitFD, atomMapD, fitAtomUnMappedL, isSkipped
        except Exception as e:
            logger.exception("%s failing for %r and %r with %s", procName, ccRefPath, molFitPath, str(e))
        return 0, {}, 0, {}, {}, [], False

    def __pairDepictPage(self, imagePath, refId, refTitle, refMol, fitId, fitTitle, fitMol, alignType="strict"):
        """Depict pairwise alignment of the input reference and fit molecules.

        Args:
            imagePath (str): path to image (format by path extension)
            refId (str): reference molecule identifier
            refTitle (str): reference molecule title
            refMol (obj): reference OE molecule object
            fitId (str): fit molecule identifier
            fitTitle (str): fit molecule title
            fitMol (obj): fit OE molecule object
            alignType (str, optional): alignment criteria (relaxed|relaxed-stereo|strict). Defaults to "strict".

        Returns:
            (list): atom mapping in all aligned figures
                    [(reference component Id, reference atom name, fit chemical component Id, fit atom name)
        """
        aML = []
        try:
            oed = OeDepictMCSAlignPage()
            oed.setSearchType(sType=alignType)

            oed.setRefMol(refMol, refId, title=refTitle)
            oed.setFitMol(fitMol, fitId, title=fitTitle)
            oed.setDisplayOptions(
                imageSizeX=2000,
                imageSizeY=1000,
                labelAtomName=True,
                labelAtomCIPStereo=True,
                labelAtomIndex=False,
                labelBondIndex=False,
                highlightStyleFit="ballAndStickInverse",
                bondDisplayWidth=0.5,
                highLightMatchColorRef="green",
                highLightNotMatchColorRef="pink",
            )
            aML = oed.alignPair(imagePath=imagePath)
            if aML:
                for (rCC, rAt, tCC, tAt) in aML:
                    logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aML

    def __depictFitList(self, sId, pdfImagePath, pairList, alignType="exact"):
        """Depict pairwise alignments with multi-page layout in PDF format.

        Args:
            pdfImagePath (str): PDF image path
            pairList (list): [(refId, refOeMol, fitId, fitOeMol)]

        Returns:
            (list): atom mapping in all aligned figures
                    [(reference component Id, reference atom name, fit chemical component Id, fit atom name)
        """
        aML = []
        try:
            logger.debug("sId %s  pairList (%d)", sId, len(pairList))
            oed = OeDepictMCSAlignMultiPage()
            oed.setSearchType(sType=alignType)
            oed.setPairMolList(pairList)

            oed.setDisplayOptions(
                labelAtomName=True,
                labelAtomCIPStereo=True,
                labelAtomIndex=False,
                labelBondIndex=False,
                highlightStyleFit="ballAndStickInverse",
                pageOrientation="portrait",
                gridRows=4,
                bondDisplayWidth=0.5,
                highLightMatchColorRef="green",
                highLightNotMatchColorRef="pink",
            )
            aML = oed.alignPairListMulti(imagePath=pdfImagePath)
            logger.debug("%s atom map length (%d)", sId, len(aML))
            if aML:
                for (rCC, rAt, tCC, tAt) in aML:
                    logger.debug("%5s %-5s %5s %-5s", rCC, rAt, tCC, tAt)
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return aML

    def __makeModelPath(self, modelDirPath, parentId, targetId, startingModelNum=1, maxModels=200, scanExisting=False):
        try:
            pth = None
            ii = startingModelNum
            dirPath = os.path.join(modelDirPath, parentId)
            if not os.access(dirPath, os.R_OK):
                os.makedirs(dirPath)
                modelId = self.__makeModelId(targetId, modelNum=ii)
                pth = os.path.join(dirPath, modelId + ".cif")
                return modelId, pth
            # optionally scan over any existing models before selecting a new model number.
            if scanExisting:
                while True:
                    modelId = self.__makeModelId(targetId, modelNum=ii)
                    pth = os.path.join(dirPath, modelId + ".cif")
                    if not os.access(pth, os.R_OK):
                        return modelId, pth
                    #
                    ii += 1
                    if ii > maxModels:
                        break
            else:
                modelId = self.__makeModelId(targetId, modelNum=ii)
                pth = os.path.join(dirPath, modelId + ".cif")
                return modelId, pth
        except Exception as e:
            logger.exception("Failing for %r %r %r with %s", parentId, targetId, modelDirPath, str(e))

        return None, None

    def __makeModelId(self, parentId, modelNum=1):
        modelId = "T_" + parentId + "_%05d" % modelNum
        return modelId

    def __getToday(self):
        """Return a CIF style date-timestamp value for current local time -"""
        today = datetime.datetime.today()
        # format ="%Y-%m-%d:%H:%M"
        fmt = "%Y-%m-%d"
        return str(today.strftime(fmt))

    def __testUnMappedProtonation(self, fitAtomUnMappedL):
        # Check for the case of extra protons ...
        try:
            fitOk = True
            if fitAtomUnMappedL:
                for atU in fitAtomUnMappedL:
                    if atU.fitAtNo != 1:
                        fitOk = False
                        break
            return fitOk
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def __writeModel(self, targetId, targetPath, fitFD, fitXyzMapD, fitAtomUnMappedL, matchObj, modelId, modelPath):
        """Write the chemical component model for the input chemical component Id and associated atom mapping and
        feature details --

            ComponentAtomDetails = namedtuple("ComponentAtomDetails", "index atNo name aType x y z fCharge")
            AlignAtomMap = namedtuple("AlignAtomMap", "refId refAtIdx refAtNo refAtName fitId fitAtIdx fitAtNo fitAtName")
            AlignAtomUnMapped = namedtuple("AlignAtomUnMapped", "fitId fitAtIdx fitAtNo fitAtType fitAtName fitAtFormalCharge x y z fitNeighbors")
        """
        try:
            unMappedTypeD = defaultdict(int)
            hAtomPrefix = "HEX"
            variantType = self.__getBuildVariant(targetId)
            #
            if not self.__testUnMappedProtonation(fitAtomUnMappedL):
                logger.info("Unmapped non-hydrogen atoms target %r model %r unMapped count (%d)", targetId, modelId, len(fitAtomUnMappedL))
                return False, variantType
            # Get atom partners for the unmapped atoms
            fitAtMapD = {}
            for refAtName, fAtTup in fitXyzMapD.items():
                fitAtMapD[fAtTup.atName] = refAtName
            if fitAtomUnMappedL:
                #  Check if neighbors are all mapped
                ok = True
                for fitUnTup in fitAtomUnMappedL:
                    for nAtName in fitUnTup.fitNeighbors:
                        if nAtName not in fitAtMapD:
                            ok = False
                            logger.info("Missing mapped neighbor for %r target %r model %r", nAtName, targetId, modelId)
                            break
                if not ok:
                    return False, variantType
                else:
                    logger.debug("%s match has unmapped protonation", modelId)
                    variantType = "tautomer_protomer"
            #
            #
            kList = ["xyz", "SMILES", "SMILES_STEREO", "InChI", "InChIKey"]
            for k in kList:
                if k not in fitFD:
                    logger.error("Fit feature dictionary for %s missing key %s", targetId, k)
                    return False, variantType
            # ------------
            dataContainer = DataContainer(modelId)
            #
            mU = MarshalUtil(workPath=self.__cachePath)
            myContainerList = mU.doImport(targetPath, fmt="mmcif")
            myContainer = myContainerList[0]
            dbName = myContainer.getName()
            if dbName.upper() != targetId.upper():
                logger.info("mismatch datablock (%r) and targetId (%r)", dbName, targetId)
            cObj = None
            if myContainer.exists("chem_comp"):
                cObj = myContainer.getObj("chem_comp")
            #
            #
            catName = "pdbx_chem_comp_model"
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["id", "comp_id"]))
            #
            parentId = targetId.split("|")[0]
            wObj = dataContainer.getObj(catName)
            wObj.setValue(modelId, "id", 0)
            wObj.setValue(parentId, "comp_id", 0)
            #
            # --------  ---------
            catName = "pdbx_chem_comp_model_atom"
            if not dataContainer.exists(catName):
                dataContainer.append(
                    DataCategory(catName, attributeNameList=["model_id", "atom_id", "type_symbol", "charge", "model_Cartn_x", "model_Cartn_y", "model_Cartn_z", "ordinal_id"])
                )
            wObj = dataContainer.getObj(catName)
            #
            if myContainer.exists("chem_comp_atom"):
                cObj = myContainer.getObj("chem_comp_atom")
            #
            #  Only write the mapped atoms in case we are missing hydrogens in the mapping
            #
            jj = 0
            for ii in range(cObj.getRowCount()):
                atName = cObj.getValue("atom_id", ii)
                atType = cObj.getValue("type_symbol", ii)
                if atName not in fitXyzMapD:
                    unMappedTypeD[atType] += 1
                    continue
                fitXyz = fitXyzMapD[atName]
                #
                # fCharge = cObj.getValue("charge", ii)
                #
                wObj.setValue(modelId, "model_id", jj)
                wObj.setValue(atName, "atom_id", jj)
                wObj.setValue(atType, "type_symbol", jj)
                #
                wObj.setValue(fitXyz.atFormalCharge, "charge", jj)
                wObj.setValue("%.4f" % fitXyz.x, "model_Cartn_x", jj)
                wObj.setValue("%.4f" % fitXyz.y, "model_Cartn_y", jj)
                wObj.setValue("%.4f" % fitXyz.z, "model_Cartn_z", jj)
                wObj.setValue(jj + 1, "ordinal_id", jj)
                jj += 1
            #
            # Add the unmapped atoms ...
            # AlignAtomUnMapped = namedtuple("AlignAtomUnMapped", "fitId fitAtIdx fitAtNo fitAtType fitAtName fitNeighbors")
            ii = wObj.getRowCount()
            for jj, uTup in enumerate(fitAtomUnMappedL):
                refAtomName = hAtomPrefix + str(jj)
                wObj.setValue(modelId, "model_id", ii)
                wObj.setValue(refAtomName, "atom_id", ii)
                wObj.setValue(uTup.fitAtType, "type_symbol", ii)
                wObj.setValue(uTup.fitAtFormalCharge, "charge", ii)
                wObj.setValue("%.4f" % uTup.x, "model_Cartn_x", ii)
                wObj.setValue("%.4f" % uTup.y, "model_Cartn_y", ii)
                wObj.setValue("%.4f" % uTup.z, "model_Cartn_z", ii)
                wObj.setValue(ii + 1, "ordinal_id", ii)
            # --------  ---------
            catName = "pdbx_chem_comp_model_bond"
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["model_id", "atom_id_1", "atom_id_2", "value_order", "ordinal_id"]))
            wObj = dataContainer.getObj(catName)
            #
            if myContainer.exists("chem_comp_bond"):
                cObj = myContainer.getObj("chem_comp_bond")
            #
            jj = 0
            for ii in range(cObj.getRowCount()):
                at1 = cObj.getValue("atom_id_1", ii)
                if at1 not in fitXyzMapD:
                    continue
                at2 = cObj.getValue("atom_id_2", ii)
                if at2 not in fitXyzMapD:
                    continue
                bType = cObj.getValue("value_order", ii)
                #
                wObj.setValue(modelId, "model_id", jj)
                wObj.setValue(at1, "atom_id_1", jj)
                wObj.setValue(at2, "atom_id_2", jj)
                wObj.setValue(bType, "value_order", jj)
                wObj.setValue(jj + 1, "ordinal_id", jj)
                jj += 1
            #
            ii = wObj.getRowCount()
            for jj, uTup in enumerate(fitAtomUnMappedL):
                at1 = hAtomPrefix + str(jj)
                for nAt in uTup.fitNeighbors:
                    at2 = fitAtMapD[nAt]
                    wObj.setValue(modelId, "model_id", ii)
                    wObj.setValue(at1, "atom_id_1", ii)
                    wObj.setValue(at2, "atom_id_2", ii)
                    wObj.setValue("SING", "value_order", ii)
                    wObj.setValue(ii + 1, "ordinal_id", ii)

            # --------  ---------
            catName = "pdbx_chem_comp_model_descriptor"
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["model_id", "type", "descriptor"]))
            wObj = dataContainer.getObj(catName)
            #
            ii = 0
            wObj.setValue(modelId, "model_id", ii)
            wObj.setValue("SMILES", "type", ii)
            wObj.setValue(fitFD["SMILES"], "descriptor", ii)
            ii += 1
            wObj.setValue(modelId, "model_id", ii)
            wObj.setValue("SMILES_CANONICAL", "type", ii)
            wObj.setValue(fitFD["SMILES_STEREO"], "descriptor", ii)
            ii += 1
            wObj.setValue(modelId, "model_id", ii)
            wObj.setValue("InChI", "type", ii)
            wObj.setValue(fitFD["InChI"], "descriptor", ii)
            ii += 1
            wObj.setValue(modelId, "model_id", ii)
            wObj.setValue("InChIKey", "type", ii)
            wObj.setValue(fitFD["InChIKey"], "descriptor", ii)
            #
            # --------  ---------
            if matchObj.getIdentifier() is not None:
                catName = "pdbx_chem_comp_model_reference"
                if not dataContainer.exists(catName):
                    dataContainer.append(DataCategory(catName, attributeNameList=["model_id", "db_name", "db_code"]))
                wObj = dataContainer.getObj(catName)
                ii = 0
                wObj.setValue(modelId, "model_id", ii)
                wObj.setValue("CSD", "db_name", ii)
                wObj.setValue(matchObj.getIdentifier(), "db_code", ii)
            #
            featureD = {}
            v = matchObj.getRFactor()
            vS = str(v)
            if v is not None and len(vS) > 0:
                featureD["r_factor"] = "%.3f" % float(v)
            #
            v = matchObj.getTemperature()
            vS = str(v)
            # remove string artifacts from temperature string ...
            if v is not None and len(vS) > 0:
                tV = vS.upper()
                try:
                    if tV.endswith("DEG.C"):
                        tV = tV.replace("AT", "")
                        tV = tV.replace("DEG.C", "")
                        tV = float(tV.strip())
                        tV = tV + 273.15
                    else:
                        tV = tV.replace("AT", "")
                        tV = tV.replace("K", "")
                        tV = float(tV.strip())
                    featureD["experiment_temperature"] = tV
                except Exception as e:
                    logger.exception("Temperature conversion fails for %s (%r) with %s", modelId, vS, tV)
            #
            v = matchObj.getCitationDOI()
            vS = str(v)
            if v is not None and len(vS) > 0:
                featureD["publication_doi"] = v
            #
            v = matchObj.getCsdVersion()
            vS = str(v)
            if v is not None and len(vS) > 0:
                featureD["csd_version"] = v
            #
            if matchObj.getRadiationSource() in ["Neutron"]:
                featureD["neutron_radiation_experiment"] = True
            if matchObj.getHasDisorder() in ["Y"]:
                featureD["has_disorder"] = True
            #
            if len(unMappedTypeD) == 1 and "H" in unMappedTypeD:
                logger.info("model %r heavy_atoms_only", modelId)
                featureD["heavy_atoms_only"] = True
            else:
                featureD["all_atoms_have_sites"] = True
            # --------  ---------
            catName = "pdbx_chem_comp_model_feature"
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["model_id", "feature_name", "feature_value"]))
            wObj = dataContainer.getObj(catName)
            #
            fKeyList = ["experiment_temperature", "publication_doi", "r_factor", "csd_version"]
            ii = 0
            for fKey in fKeyList:
                if fKey in featureD:
                    wObj.setValue(modelId, "model_id", ii)
                    wObj.setValue(fKey, "feature_name", ii)
                    wObj.setValue(str(featureD[fKey]), "feature_value", ii)
                    ii += 1

            #
            boolKeyList = ["has_disorder", "neutron_radiation_experiment", "heavy_atoms_only", "all_atoms_have_sites"]
            for fKey in boolKeyList:
                if fKey in featureD:
                    if featureD[fKey]:
                        wObj.setValue(modelId, "model_id", ii)
                        wObj.setValue(fKey, "feature_name", ii)
                        wObj.setValue("Y", "feature_value", ii)
                        ii += 1
            #

            if variantType:
                wObj.setValue(modelId, "model_id", ii)
                wObj.setValue(variantType + "_match", "feature_name", ii)
                wObj.setValue("Y", "feature_value", ii)
                ii += 1

            # --------  ---------
            catName = "pdbx_chem_comp_model_audit"
            if not dataContainer.exists(catName):
                dataContainer.append(DataCategory(catName, attributeNameList=["model_id", "action_type", "date"]))
            wObj = dataContainer.getObj(catName)
            #
            ii = 0
            wObj.setValue(modelId, "model_id", ii)
            wObj.setValue("Initial release", "action_type", ii)
            wObj.setValue(self.__getToday(), "date", ii)
            # wObj.setValue('RCSB', 'processing_site',  ii)
            # wObj.setValue('JDW', 'annotator', ii)
            # wObj.setValue('?', 'details', ii)
            #
            ok = mU.doExport(modelPath, [dataContainer], fmt="mmcif")
            return ok, variantType
        except Exception as e:
            logger.exception("Failing for %r %r with %s", targetId, targetPath, str(e))
        return False, ""


class ChemCompModelBuild(object):
    def __init__(self, cachePath, prefix=None):
        self.__cachePath = cachePath

        self.__prefix = prefix
        startTime = time.time()
        useCache = True
        ccFileNamePrefix = "cc-%s" % self.__prefix if self.__prefix else "cc"
        self.__ccSIdxP = ChemCompSearchIndexProvider(cachePath=cachePath, useCache=useCache, ccFileNamePrefix=ccFileNamePrefix)
        ok = self.__ccSIdxP.testCache()
        logger.info("Completed chemical component search index load %r (%.4f seconds)", ok, time.time() - startTime)
        #
        self.__startTime = time.time()
        logger.info("Starting model build (%s) at %s", __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def getModelDirFilePath(self):
        dN = "cc-%s-model-files" % self.__prefix if self.__prefix else "cc-model-files"
        return os.path.join(self.__cachePath, dN)

    def getModelImageDirFilePath(self):
        dN = "cc-%s-model-image" % self.__prefix if self.__prefix else "cc-model-images"
        return os.path.join(self.__cachePath, dN)

    def fetchModelIndex(self):
        mD = {}
        try:
            mU = MarshalUtil(workPath=self.__cachePath)
            fp = self.__getModelIndexPath()
            mD = mU.doImport(fp, fmt="json")
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return mD

    def storeModelIndex(self, mD):
        try:
            mU = MarshalUtil(workPath=self.__cachePath)
            fp = self.__getModelIndexPath()
            ok = mU.doExport(fp, mD, fmt="json", indent=3)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            ok = False
        return ok

    def __getModelIndexPath(self):
        return os.path.join(self.getModelDirFilePath(), "model-index.json")

    def build(self, alignType="relaxed-stereo", numProc=4, chunkSize=10, verbose=False):
        """Run the model build step in the chemical component model workflow.

        Args:
          alignType (str):  "relaxed"|"strict"| relaxed-stereo".  Default: relaxed-stereo
          numProc (int, optional): number of processes to invoke. Defaults to 4.
          chunkSize (int, optional): work chunksize. Defaults to 10.
          verbose (bool, optional): verbose logging.  Defaults to False.

        Returns:
            (dict): {searchId: [{"targetId": , "modelId": , "modelPath": ,"matchId": , "parentId": , "rFactor": , }]

        """
        retD = {}
        try:
            ccms = ChemCompModelSearch(self.__cachePath, None, None, prefix=self.__prefix)
            modelDirPath = self.getModelDirFilePath()
            imageDirPath = self.getModelImageDirFilePath()
            #
            idxPathD = ccms.getResultIndex()
            idxPathL = list(idxPathD.values())
            pD = {}
            for sId in idxPathD:
                parentId = sId.split("|")[0]
                pD.setdefault(parentId, []).append(sId)
            logger.info("Using search result index length ridxD (%d) parent coverage (%d)", len(idxPathD), len(pD))
            #
            pU = ChemCompModelBuildWorker(self.__cachePath, verbose=verbose)
            mpu = MultiProcUtil(verbose=True)
            mpu.setWorkingDir(modelDirPath)
            mpu.setOptions(optionsD={"modelDirPath": modelDirPath, "imageDirPath": imageDirPath, "alignType": alignType, "ccSIdxP": self.__ccSIdxP})
            #
            mpu.set(workerObj=pU, workerMethod="build")

            ok, failList, resultList, _ = mpu.runMulti(dataList=idxPathL, numProc=numProc, numResults=1, chunkSize=chunkSize)
            logger.info("Run ended with status %r success count %d failures %r", ok, len(resultList[0]), len(failList))
            successList = copy.copy(resultList[0])
            for tD in successList:
                retD.setdefault(tD["parentId"], []).append(tD)
            #
            if retD:
                logger.info("Completed build with models for %d parent chemical definitions", len(retD))
            else:
                logger.info("No models built")
            ok = self.storeModelIndex(retD)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return retD



##
# File: ChemCompModelExec.py
# Date: 26-Jan-2021  jdw
#
#  Execution wrapper  --  for chemical component model construction workflow
#
#  Updates:
#
##
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import argparse
import logging
import os
import sys

from rcsb.ccmodels.search import __version__
from rcsb.ccmodels.search.ChemCompModelAssemble import ChemCompModelAssemble
from rcsb.ccmodels.search.ChemCompModelBuild import ChemCompModelBuild
from rcsb.ccmodels.search.ChemCompModelGen import ChemCompModelGen
from rcsb.ccmodels.search.ChemCompModelSearch import ChemCompModelSearch

# from rcsb.utils.io.MarshalUtil import MarshalUtil

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser()
    #
    parser.add_argument("--generate", default=False, action="store_true", help="Generate searchable files")
    parser.add_argument("--cc_locator", default=None, help="Chemical component reference dictioanary locator")
    parser.add_argument("--bird_locator", default=None, help="BIRD reference dictioanary locator")
    parser.add_argument("--prefix", default=None, help="Prefix for identifying reference resource files (e.g. abbrev)")
    parser.add_argument("--cache_path", default=None, help="Top-level cache directory path")
    parser.add_argument("--use_cache", default=False, action="store_true", help="Re-use cached resource files")
    parser.add_argument("--limit_perceptions", default=False, action="store_true", help="Restrict automatic OE chemical perceptions")
    #
    parser.add_argument("--search", default=False, action="store_true", help="Execute CCDC search")
    parser.add_argument("--update_only", default=False, action="store_true", help="Only update current search results")
    #
    parser.add_argument("--build", default=False, action="store_true", help="Build models from CCDC search results")
    parser.add_argument("--build_align_type", default=False, action=None, help="Alignment criteria (default: graph-relaxed-stereo-sdeq")
    #
    parser.add_argument("--assemble", default=False, action="store_true", help="Assemble models into a concatenated file")
    parser.add_argument("--max_r_factor", default=10.0, help="Maximum permissible R-value in assembled model file (default=10.0)")
    #
    parser.add_argument("--csdhome", default=None, help="Path to the CSD release (path to CSD_202x)")
    parser.add_argument("--python_lib_path", default=None, help="Path to Python library")
    parser.add_argument("--python_version", default=None, help="Python library version (default: 3.7)")
    #
    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--search_timeout", default=240, help="Search timeout (seconds) default=240")

    parser.add_argument("--verbose", default=False, action="store_true", help="Verbose output")

    args = parser.parse_args()
    #
    try:
        pyLib = args.python_lib_path if args.python_lib_path else os.path.join(os.environ["PYENV_ROOT"], "versions", "3.7.9", "lib")
        pyRoot = os.path.dirname(pyLib)
        pyVer = args.python_version if args.python_version else "3.7"
        csdHome = args.csdhome if args.csdhome else os.environ["CSDHOME"]
        #
        doGen = args.generate
        ccLocator = args.cc_locator
        birdLocator = args.bird_locator
        prefix = args.prefix
        cachePath = args.cache_path
        numProc = int(args.num_proc)
        chunkSize = int(args.chunk_size)
        useCache = args.use_cache
        limitPerceptions = args.limit_perceptions
        #
        doSearch = args.search
        searchType = "substructure"
        updateOnly = args.update_only
        #
        doBuild = args.build
        doAssemble = args.assemble
        maxRFactor = args.max_r_factor
        verbose = args.verbose
        alignType = args.build_align_type if args.build_align_type else "graph-relaxed-stereo-sdeq"
        searchTimeOut = int(args.search_timeout)
    except Exception as e:
        logger.exception("Argument processing problem %s", str(e))
        parser.print_help(sys.stderr)
        exit(1)
    #
    try:
        os.environ["CSDHOME"] = csdHome
        os.environ["LD_LIBRARY_PATH"] = "%s:%s/python%s/site-packages/ccdc/_lib:$LD_LIBRARY_PATH" % (pyLib, pyLib, pyVer)
        os.environ["DYLD_LIBRARY_PATH"] = "%s/python%s/site-packages/ccdc/_lib" % (pyLib, pyVer)
        os.environ["DYLD_FRAMEWORK_PATH"] = "%s/python%s/site-packages/ccdc/_lib" % (pyLib, pyVer)

        logger.info("Using CSDHOME %s", os.environ["CSDHOME"])
        logger.info("Using DYLD_LIBRARY_PATH %s", os.environ["DYLD_LIBRARY_PATH"])
        logger.info("Using DYLD_FRAMEWORK_PATH %s", os.environ["DYLD_FRAMEWORK_PATH"])
        logger.info("Using source version %s", __version__)
        # -----------
        if doGen:
            ccmG = ChemCompModelGen(cachePath=cachePath, prefix=prefix)
            numMols = ccmG.buildSearchFiles(
                ccUrlTarget=ccLocator,
                birdUrlTarget=birdLocator,
                numProc=numProc,
                chunkSize=chunkSize,
                minCount=30,
                useCache=useCache,
                limitPerceptions=limitPerceptions,
            )
            logger.info("Generated %d search files using prefix %r", numMols, prefix)

        if doSearch:
            ccms = ChemCompModelSearch(cachePath=cachePath, pythonRootPath=pyRoot, csdHome=csdHome, prefix=prefix)
            rL = ccms.search(searchType, updateOnly=updateOnly, numProc=numProc, chunkSize=chunkSize, timeOut=searchTimeOut)
            logger.info("Search success count %d", len(rL))

        if doBuild:
            ccmb = ChemCompModelBuild(cachePath=cachePath, prefix=prefix)
            rD = ccmb.build(alignType=alignType, numProc=numProc, chunkSize=chunkSize, verbose=verbose)
            logger.info("Built model count %d", len(rD))

        if doAssemble:
            ccmb = ChemCompModelAssemble(cachePath=cachePath, prefix=prefix)
            numAssem = ccmb.assemble(maxRFactor=maxRFactor)
            logger.info("Assembled model count %d", numAssem)

    except Exception as e:
        logger.exception("Failing with %s", str(e))

    # ----------------------- - ----------------------- - ----------------------- - ----------------------- - ----------------------- -


if __name__ == "__main__":
    main()



##
#
# File:    testChemCompModelSearch.py
# Author:  J. Westbrook
# Date:    16-Jan-2021
# Version: 0.001
#
# Updated:
#
##
"""
Test cases for CCDC search step in chemical component model workflow.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import unittest
import time
import os
import os.path
import platform
import resource

from rcsb.ccmodels.search.ChemCompModelSearch import ChemCompModelSearch

from rcsb.ccmodels.search import __version__


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ChemCompModelSearchTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__prefix = "abbrev"
        self.__numProc = 2
        self.__chunkSize = 5
        self.__pythonRootPath = os.path.join(os.environ["CSD_PYTHON_ROOT_PATH"])
        self.__csdHome = os.environ["CSDHOME"]
        #
        self.__searchType = "substructure"
        #
        self.__startTime = time.time()
        logger.info("Starting %s (%s) at %s", self.id(), __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testSearchWorkflow(self):
        """Test case: CCDC substructure search workflow step"""
        try:
            ccms = ChemCompModelSearch(cachePath=self.__cachePath, pythonRootPath=self.__pythonRootPath, csdHome=self.__csdHome, prefix=self.__prefix)
            rL = ccms.search(self.__searchType, updateOnly=False, numProc=self.__numProc, chunkSize=self.__chunkSize, timeOut=360)
            logger.info("Search success list (%d) %r", len(rL), rL)
            self.assertGreaterEqual(len(rL), 18)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()

    @unittest.skipIf(skipFlag, "Updte search targets - troubleshooting test")
    def testSearchUpdateWorkflow(self):
        """Test case: CCDC substructure search update workflow step"""
        try:
            ccms = ChemCompModelSearch(cachePath=self.__cachePath, pythonRootPath=self.__pythonRootPath, csdHome=self.__csdHome, prefix=self.__prefix)
            rL = ccms.search(self.__searchType, updateOnly=True, numProc=self.__numProc, chunkSize=self.__chunkSize, timeOut=360)
            logger.info("Update search success list (%d) %r", len(rL), rL)
            self.assertGreaterEqual(len(rL), 0)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def suiteSearchTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ChemCompModelSearchTests("testSearchWorkflow"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite = suiteSearchTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    ChemCompModelSearch.py
# Author:  J. Westbrook
# Date:    12-Jan-2021
# Version: 0.001
#
# Updated:
##
"""
CCDC search step in chemical component model workflow.


"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import glob
import logging
import time
import os

from rcsb.ccmodels.search.CcdcSearchExecMp import CcdcSearchExecMp
from rcsb.ccmodels.search.ChemCompModelGen import ChemCompModelGen
from rcsb.ccmodels.search import __version__

# from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompModelSearch(object):
    def __init__(self, cachePath, pythonRootPath, csdHome, prefix=None):
        self.__cachePath = cachePath
        self.__pythonRootPath = pythonRootPath
        self.__csdHome = csdHome
        self.__prefix = prefix
        self.__startTime = time.time()
        logger.info("Starting search (%s) at %s", __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def getResultDirFilePath(self):
        dN = "cc-%s-result-files" % self.__prefix if self.__prefix else "cc-result-files"
        return os.path.join(self.__cachePath, dN)

    def search(self, searchType, updateOnly=False, numProc=2, chunkSize=10, fileType="sdf", timeOut=240):
        """Run the CCDC search step in chemical component model workflow.

        Args:
            searchType (str): search type: substructure|similarity
            updateOnly (bool, optional): only update search results. Defaults to False.
            numProc (int, optional): number of processors to invoke. Defaults to 2.
            chunkSize (int, optional): incremental chunk size for each subprocess. Defaults to 10.
            fileType (str, optional): molecule file format (sdf|mol2) default: sdf
            timeOut (int, optional): search time out (seconds).  defaults 240

        Returns:
            (list): searchId list for successful searches
        """
        rL = []
        try:
            resultDirPath = self.getResultDirFilePath()
            ccmg = ChemCompModelGen(self.__cachePath, prefix=self.__prefix)
            # sourceSearchDirPath = ccmg.getSearchDirFilePath()
            sTupL = ccmg.fetchPathList()
            # sTupL = self.__getSearchPathListGlob(sourceSearchDirPath, fileType=fileType)
            if updateOnly:
                rD = self.getResultIndex()
                pL = [sTup[1] for sTup in sTupL if sTup[0] not in rD and sTup[2] == fileType]
                logger.info("Updated search list length %d", len(pL))
            else:
                pL = [sTup[1] for sTup in sTupL if sTup[2] == fileType]

            if pL:
                csmp = CcdcSearchExecMp(pythonRootPath=self.__pythonRootPath, csdHome=self.__csdHome)
                resultL = csmp.runSearch(pL, resultDirPath, searchType=searchType, numProc=numProc, chunkSize=chunkSize, timeOut=timeOut)
                for result in resultL:
                    bfn, _ = os.path.splitext(os.path.basename(result))
                    rL.append(bfn)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return rL

    def getResultIndex(self):
        """Return the dictionary index of search results {identifier: idx file path}

        Args:
            resultDirPath (str): directory containing searchable mol2 files

        Returns:
            (dict): {searchId: search index file path}

        """
        rD = {}
        try:
            resultDirPath = self.getResultDirFilePath()
            logger.info("Using result directory path %r", resultDirPath)
            pthL = glob.glob(os.path.join(resultDirPath, "**", "*-index.json"), recursive=True)
            logger.info("Result path list length %d", len(pthL))
            #
            # Parse out the identifier list -
            #
            for pth in pthL:
                bfn, _ = os.path.splitext(os.path.basename(pth))
                sId = bfn.replace("-index", "")
                rD[sId] = pth
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return rD

    def __getSearchPathListGlob(self, sourceSearchDirPath, fileType="sdf"):
        """Return the path list of mol files in the input path

        Args:
            sourceSearchDirPath (str): directory containing searchable files
            fileType (str): mol2 or sdf default: sdf

        Returns:
            (list of tuples): [(searchId, molfile path)]
        """
        tupL = []
        try:
            logger.info("Using mol2 directory path %r", sourceSearchDirPath)
            fPattern = "*." + fileType
            pthL = glob.glob(os.path.join(sourceSearchDirPath, "**", fPattern), recursive=True)
            logger.info("Search list length %d", len(pthL))
            #
            # Parse out the identifier list -
            #
            for pth in pthL:
                bfn, _ = os.path.splitext(os.path.basename(pth))
                tupL.append((bfn, pth))
            #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return tupL



##
#
# File:    testChemCompModelAssemble.py
# Author:  J. Westbrook
# Date:    4-Feb-2021
# Version: 0.001
#
# Updated:
#
##
"""
Test cases for to assemble models from CCDC search results in the chemical component model workflow.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import unittest
import time
import os
import os.path
import platform
import resource

from rcsb.ccmodels.search.ChemCompModelAssemble import ChemCompModelAssemble

from rcsb.ccmodels.search import __version__


HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()
logger.setLevel(logging.INFO)


class ChemCompModelAssembleTests(unittest.TestCase):
    def setUp(self):
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__urlTarget = os.path.join(HERE, "test-data", "chem_comp_models.cif")
        self.__prefix = "abbrev"
        self.__numProc = 2
        self.__chunkSize = 5
        #
        self.__startTime = time.time()
        logger.info("Starting %s (%s) at %s", self.id(), __version__, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testAssembleWorkflow(self):
        """Test case: model assemble workflow step"""
        try:
            ccma = ChemCompModelAssemble(cachePath=self.__cachePath, prefix=self.__prefix, urlTarget=self.__urlTarget)
            numC = ccma.assemble(maxRFactor=10.0)
            self.assertGreaterEqual(numC, 131)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
            self.fail()


def suiteAssembleTests():
    suiteSelect = unittest.TestSuite()
    suiteSelect.addTest(ChemCompModelAssembleTests("testAssembleWorkflow"))
    return suiteSelect


if __name__ == "__main__":
    #
    mySuite = suiteAssembleTests()
    unittest.TextTestRunner(verbosity=2).run(mySuite)



##
# File:    ChemCompModelGenTests.py
# Author:  J. Westbrook
# Date:    12-Jan-2021
# Version: 0.001
#
# Update:
#
#
##
"""
Tests utilities building CCDC searchable source files for chemical component models.
"""

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os
import platform
import resource
import time
import unittest

from rcsb.ccmodels.search import __version__
from rcsb.ccmodels.search.ChemCompModelGen import ChemCompModelGen

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
logger = logging.getLogger()


class ChemCompModelGenTests(unittest.TestCase):
    skipFlag = True

    def setUp(self):
        self.__startTime = time.time()
        self.__dataPath = os.path.join(HERE, "test-data")
        self.__cachePath = os.path.join(HERE, "test-output", "CACHE")
        self.__ccUrlTarget = os.path.join(self.__dataPath, "components-abbrev.cif")
        self.__birdUrlTarget = os.path.join(self.__dataPath, "prdcc-abbrev.cif")
        #
        logger.debug("Running tests on version %s", __version__)
        logger.info("Starting %s at %s", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()))

    def tearDown(self):
        unitS = "MB" if platform.system() == "Darwin" else "GB"
        rusageMax = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
        logger.info("Maximum resident memory size %.4f %s", rusageMax / 10 ** 6, unitS)
        endTime = time.time()
        logger.info("Completed %s at %s (%.4f seconds)", self.id(), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - self.__startTime)

    def testMoleculeCacheFilesAbbrev(self):
        ccmG = ChemCompModelGen(cachePath=self.__cachePath, prefix="abbrev")
        numMols = ccmG.buildSearchFiles(
            ccUrlTarget=self.__ccUrlTarget,
            birdUrlTarget=self.__birdUrlTarget,
            numProc=2,
            minCount=30,
            useCache=False,
            limitPerceptions=False,
        )
        # For 156 limitPerceptions=False  and 160 for limitPerceptions=True
        self.assertGreaterEqual(numMols, 155)

    @unittest.skipIf(skipFlag, "Build search targets for the full dictionary - troubleshooting test")
    def testBuildMoleculeCacheFilesFull(self):
        ccmG = ChemCompModelGen(cachePath=self.__cachePath)
        ccmG.buildSearchFiles(
            molLimit=None,
            minCount=500,
            fpTypeList=None,
            screenTypeList=None,
            numProc=2,
            useCache=False,
            limitPerceptions=False,
        )



##
# File:    ChemCompModelGen.py
# Author:  J. Westbrook
# Date:    12-Jan-2021
# Version: 0.001
#
# Updated:
#  27-Jan-2021 jdw add __checkCif() and __checkOeMol() to filter molecules not subject to search
##
"""
Generate searchable source files for chemical component model workflow.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

import logging
import time
import os

from rcsb.utils.chem.ChemCompMoleculeProvider import ChemCompMoleculeProvider
from rcsb.utils.chem.ChemCompSearchIndexProvider import ChemCompSearchIndexProvider
from rcsb.utils.chem.OeChemCompUtils import OeChemCompUtils
from rcsb.utils.chem.OeIoUtils import OeIoUtils
from rcsb.utils.chem.OeSearchMoleculeProvider import OeSearchMoleculeProvider
from rcsb.utils.io.MarshalUtil import MarshalUtil

logger = logging.getLogger(__name__)


class ChemCompModelGen(object):
    def __init__(self, cachePath, prefix=None):
        self.__cachePath = cachePath
        self.__prefix = prefix if prefix else ""

    def getSearchDirFilePath(self):
        dN = "cc-%s-search-files" % self.__prefix if self.__prefix else "cc-search-files"
        return os.path.join(self.__cachePath, dN)

    def getIndexFilePath(self):
        dN = "cc-%s-search-files" % self.__prefix if self.__prefix else "cc-search-files"
        return os.path.join(self.__cachePath, dN, "search-file-index.json")

    def buildSearchFiles(self, **kwargs):
        """Build cif, sdf (optional), and mol2 files for components in the chemical component search index.
           Exclude ions or other extraneous molecules lacking bonds.

        Args:
            ccUrlTarget (str): locator for source chemical component dictionary (default: full public dictionary)
            birdUrlTarget (str): locator for source BIRD dictionary (default: full public dictionary)
            limitPerceptions (bool): restrict automatic perceptions in OE molecular build operations (default: False)
            numProc (int): number of processors
            useCache (bool): use existing resource file where possible (default: True)
            molLimit (str):  limit the number to ingested chemical compont (default: None)
            quietFlag (bool): suppress output in OE library operations (default: True)

        Returns:
            (int): number molfiles generated
        """
        cachePath = self.__cachePath
        ccUrlTarget = kwargs.get("ccUrlTarget", None)
        birdUrlTarget = kwargs.get("birdUrlTarget", None)
        molLimit = kwargs.get("molLimit", None)
        quietFlag = kwargs.get("quietFlag", True)
        fpTypeList = kwargs.get("fpTypeList", [])
        screenTypeList = kwargs.get("screenTypeList", [])
        ccFileNamePrefix = "cc-%s" % self.__prefix if self.__prefix else "cc"
        oeFileNamePrefix = "oe-%s" % self.__prefix if self.__prefix else "oe"
        numProc = kwargs.get("numProc", 2)
        minCount = kwargs.get("minCount", 0)
        useCache = kwargs.get("useCache", True)
        useSdf = kwargs.get("useSdf", True)
        useMol2 = kwargs.get("useMol2", False)
        limitPerceptions = kwargs.get("limitPerceptions", False)
        logSizes = False
        #
        startTime = time.time()
        ccmP = ChemCompMoleculeProvider(
            cachePath=cachePath, useCache=useCache, ccFileNamePrefix=ccFileNamePrefix, ccUrlTarget=ccUrlTarget, birdUrlTarget=birdUrlTarget, molLimit=molLimit
        )
        ok = ccmP.testCache(minCount=minCount, logSizes=logSizes)
        logger.info("Completed chemical component provider load %r (%.4f seconds)", ok, time.time() - startTime)
        #
        startTime = time.time()
        oesmp = OeSearchMoleculeProvider(
            ccUrlTarget=ccUrlTarget,
            birdUrlTarget=birdUrlTarget,
            cachePath=cachePath,
            ccFileNamePrefix=ccFileNamePrefix,
            oeFileNamePrefix=oeFileNamePrefix,
            useCache=useCache,
            quietFlag=quietFlag,
            fpTypeList=fpTypeList,
            screenTypeList=screenTypeList,
            numProc=numProc,
            molLimit=molLimit,
            limitPerceptions=limitPerceptions,
        )
        ok = oesmp.testCache()
        logger.info("Completed OE molecule provider load %r (%.4f seconds)", ok, time.time() - startTime)
        #
        startTime = time.time()
        ccSIdxP = ChemCompSearchIndexProvider(cachePath=cachePath, useCache=useCache, ccFileNamePrefix=ccFileNamePrefix, limitPerceptions=limitPerceptions, numProc=numProc)
        ok = ccSIdxP.testCache()
        logger.info("Completed chemical component search index load %r (%.4f seconds)", ok, time.time() - startTime)
        #
        ccSIdx = ccSIdxP.getIndex() if ccSIdxP and ok else {}
        logger.info("Search index status %r index length %d", ok, len(ccSIdx))
        #
        ccIdD = {}
        mU = MarshalUtil()
        oeU = OeIoUtils(dirPath=cachePath)
        numMols = 0
        searchFileDirPath = self.getSearchDirFilePath()
        pathTupList = []
        for sId in ccSIdx:
            ccId = sId.split("|")[0]
            # standard CIF definition
            if ccId not in ccIdD:
                cifPath = os.path.join(searchFileDirPath, ccId[0], ccId, ccId + ".cif")
                if not (useCache and mU.exists(cifPath)):
                    ccMol = ccmP.getMol(ccId)
                    if not self.__checkCif(ccMol):
                        continue
                    mU.doExport(cifPath, [ccMol], fmt="mmcif")
            #
            oeMol = oesmp.getMol(sId)
            if not self.__checkOeMol(oeMol):
                continue
            #
            # Sanity checks on the generated OE molecule
            #
            cifPath = os.path.join(searchFileDirPath, ccId[0], ccId, sId + ".cif")
            if sId != ccId and not (useCache and mU.exists(cifPath)):
                oeccU = OeChemCompUtils()
                ok = oeccU.addOeMol(sId, oeMol, missingModelXyz=True, writeIdealXyz=False)
                if ok:
                    oeccU.write(cifPath)

            if useSdf:
                molFilePath = os.path.join(searchFileDirPath, ccId[0], ccId, sId + ".sdf")
                if not (useCache and mU.exists(molFilePath)):
                    ok = oeU.write(molFilePath, oeMol, constantMol=False, addSdTags=True)
                    if ok:
                        pathTupList.append((sId, molFilePath, "sdf"))
            #
            if useMol2:
                mol2FilePath = os.path.join(searchFileDirPath, ccId[0], ccId, sId + ".mol2")
                if not (useCache and mU.exists(mol2FilePath)):
                    oeU.write(mol2FilePath, oeMol, constantMol=False, addSdTags=True)
                    if ok:
                        pathTupList.append((sId, mol2FilePath, "mol2"))
            numMols += 1
        #
        self.__storePathList(pathTupList)
        return numMols

    def fetchPathList(self):
        pathList = []
        fp = self.getIndexFilePath()
        try:
            mU = MarshalUtil()
            pathList = mU.doImport(fp, fmt="json")
        except Exception as e:
            logger.info("Failing for %r with %s", fp, str(e))
        return pathList

    def __storePathList(self, pathList):
        ok = False
        fp = self.getIndexFilePath()
        try:
            mU = MarshalUtil()
            ok = mU.doExport(fp, pathList, fmt="json", indent=3)
        except Exception as e:
            logger.info("Failing for %r with %s", fp, str(e))
        return ok

    def __checkOeMol(self, oeMol):
        """Test if input OEmol is sufficiently complete to be included as a search target.

        Args:
            oeMol (obj): OE molecule object

        Returns:
           (bool): True for success or False otherwise

        """
        numAtoms = numBonds = 0
        # Contains multiple atoms -
        try:
            numAtoms = oeMol.NumAtoms()
        except Exception:
            return False

        # Contains bonds -
        try:
            numBonds = oeMol.NumBonds()
        except Exception:
            return False

        return numAtoms > 1 and numBonds > 0

    def __checkCif(self, dataContainer):
        """Test if input CIF is sufficiently complete to be included as a search target.

        Args:
            dataContainer (obj): data container holding chemical component definition

        Returns:
           (bool): True for success or False otherwise

        """
        # Contains > 1 atoms -
        try:
            cObj = dataContainer.getObj("chem_comp_atom")
            numAtoms = cObj.getRowCount()
        except Exception:
            return False

        # Contains > 0 bonds -
        try:
            cObj = dataContainer.getObj("chem_comp_bond")
            numBonds = cObj.getRowCount()
        except Exception:
            return False

        return numAtoms > 1 and numBonds > 0



##
#
# File:    CcdcSearchExecMp.py
# Author:  J. Westbrook
# Date:    15-Jan-2021
# Version: 0.001
#
# Updated:
#
##
"""
Multiprocessing wrapper for substructure and similarity search against the CCDC local Python API -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "john.westbrook@rcsb.org"
__license__ = "Apache 2.0"

# pylint: disable=redefined-outer-name

import copy
import logging
import time
import os
import os.path

from rcsb.utils.io.ExecUtils import ExecUtils
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.multiproc.MultiProcUtil import MultiProcUtil

logger = logging.getLogger(__name__)


class CcdcSearchExecWorker(object):
    def __init__(self, verbose=True):
        self.__verbose = verbose

    def __checkStop(self, path):
        try:
            if os.access(path, os.F_OK):
                return True
        except Exception:
            pass
        return False

    def search(self, dataList, procName, optionsD, workingDir):
        """Worker method to execute a shell to search CCDC for the input mol2 path list.

        Args:
            dataList (list): list of mol2 file paths to be searched
            procName (str): processName
            optionsD (dict): dictionary of options
            workingDir (str): path to working directory (not used)

        Returns:
            (successList, resultList, []): success and result lists of mol2 paths with CCDC matches
        """
        resultPath = optionsD["resultPath"]
        searchType = optionsD["searchType"]
        pythonRootPath = optionsD["pythonRootPath"]
        csdHome = optionsD["csdHome"]
        timeOut = optionsD["timeOut"]
        timeOut = timeOut if timeOut and timeOut > 0 else 120
        _ = workingDir
        resultList = []
        startTime = time.time()
        logger.info("starting %s at %s", procName, time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        #
        try:
            stopPath = os.path.join(resultPath, "STOP")
            logger.info("%s starting search data length %d", procName, len(dataList))
            if self.__checkStop(stopPath):
                logger.info("%s stopping", procName)
                return resultList, resultList, []
            #
            queryListFilePath = os.path.join(resultPath, procName, "queryFileList.list")
            mU = MarshalUtil()
            ok = mU.doExport(queryListFilePath, dataList, fmt="list")
            if not ok:
                return resultList, resultList, []
            #
            exU = ExecUtils()
            logger.debug("%s executing shell for %s", procName, queryListFilePath)
            cmdPath = os.path.join(pythonRootPath, "bin", "ccdc_search_cli")
            hitListPath = os.path.join(resultPath, procName, "hitList.list")
            logPath = os.path.join(resultPath, procName, "execlog.log")

            logger.debug("cmdPath %r", cmdPath)
            ok = exU.runShell(
                "%s --mol_list_path %s --result_path %s --search_type %s --csdhome %s --hit_list_path %s"
                % (cmdPath, queryListFilePath, resultPath, searchType, csdHome, hitListPath),
                outPath=logPath,
                outAppend=True,
                timeOut=timeOut,
                suppressStderr=False,
            )
            #
            if ok and mU.exists(hitListPath):
                resultList = mU.doImport(hitListPath, fmt="list")
        except Exception as e:
            logger.exception("Failing with %s", str(e))

        endTime = time.time()
        logger.info("%s (result length %d) completed at %s (%.2f seconds)", procName, len(resultList), time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return resultList, resultList, []


class CcdcSearchExecMp(object):
    def __init__(self, pythonRootPath, csdHome, verbose=True):
        """MP execution wrapper for variant search against the CCDC local Python API -"""
        self.__verbose = verbose
        self.__pythonRootPath = pythonRootPath
        self.__csdHome = csdHome
        #

    def runSearch(self, molFilePathList, resultPath, searchType="similarity", numProc=4, chunkSize=10, timeOut=120):
        """Run CCDC search in multiprocess mode.

        Args:
            molFilePathList (list): input mol2/sdf path list to search
            resultPath (str): directory path to store results
            searchType (str, optional): search type (substructure|similarity). Defaults to "similarity".
            numProc (int, optional): number of processes to invoke. Defaults to 4.
            chunkSize (int, optional): work chunksize. Defaults to 10.
            timeOut (int, optional): search timeout Defaults: 120 seconds.
        """
        logger.info("Starting with molfile path list length %d", len(molFilePathList))
        successList = []
        try:
            pU = CcdcSearchExecWorker(verbose=self.__verbose)
            mpu = MultiProcUtil(verbose=True)
            mpu.setWorkingDir(resultPath)
            mpu.setOptions(optionsD={"resultPath": resultPath, "searchType": searchType, "pythonRootPath": self.__pythonRootPath, "csdHome": self.__csdHome, "timeOut": timeOut})
            #
            mpu.set(workerObj=pU, workerMethod="search")

            ok, failList, resultList, _ = mpu.runMulti(dataList=molFilePathList, numProc=numProc, numResults=1, chunkSize=chunkSize)
            logger.info("Run ended with status %r success count %d failures %r", ok, len(resultList[0]), len(failList))
            successList = copy.copy(resultList[0])
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return successList
